DEFAULT_QUEUE_NAME = "streaq"
DEFAULT_TTL = 86_400_000  # 1 day in ms
REDIS_ABORT = ":aborted"
REDIS_CHANNEL = ":channel"
REDIS_GROUP = "workers"
REDIS_HEALTH = ":health"
REDIS_LOCK = ":lock"
REDIS_MESSAGE = ":task:messages:"
REDIS_RESULT = ":task:results:"
REDIS_RETRY = ":task:retry:"
REDIS_RUNNING = ":task:running:"
REDIS_QUEUE = ":queue"
REDIS_STREAM = ":stream"
REDIS_TASK = ":task:data:"
REDIS_TIMEOUT = ":timeout"
